# ------------------------------------------------------------------
#
# $Id: lock.pm,v 1.8 2007-10-15 23:35:19 rodney Exp $
#

package rvlib::lock;

use strict;
use warnings;

use IO::Handle;
use Fcntl ':flock'; # import LOCK_* constants
use Time::HiRes qw( usleep time );

use vars qw(@ISA);

#
# Usage:
#
# use rvlib::lock;
#
# my $lock = new rvlib::lock(lockdir => '/tmp/mylock');
#

sub new {
    my ($this,%c) = @_;
    my $class = ref ($this) || $this; # class or object
    my $self = {};

    # "this" ist der Klassenname,
    # "self" ist der Instanzname.

    bless($self, $class);

    $self->init(%c);

    return($self);
}



sub init {
    my ($this,%c) = @_;
    my $c = \%c;

#    $self->init ();


    if ($c->{lockdir}) {
	$this->{lockdir} = $c->{lockdir};
    }
    else {
	die("missing lockdir => 'filename' in new rvlib::lock(). Terminating");
    }

    # Use default, if no lock timeout given
    $this->{locktimeout} = $c->{locktimeout} || 5;

    # Retry time in msecs
    $this->{lockretrytime} = $c->{lockretrytime} || 300;

    $this->{lockslots} = $c->{lockslots} || 1;

    $this->{debug} = $c->{debug} || 0;

    if ($c->{logfile}) {
	$this->{logfile} = $c->{logfile};
    }

    # Verzeichnis anlegen, falls es noch nicht existiert
    if (! -d $this->{lockdir}) {
	mkdir($this->{lockdir},0775);
    }

    if (! -d $this->{lockdir}) {
	die("could not create lock dir $this->{lockdir}. Terminated");
    }


    # Get a lock

    $this->{lockslot} = undef;

    my $t0 = time();

    # Find a free lock slot
    do {
	for (my $n = 0; $n < $this->{lockslots}; $n++) {
	    my $file = sprintf("%s/%03d",$this->{lockdir},$n);
	    
	    $this->{lockhandle} = new IO::Handle;
	    open($this->{lockhandle},"+>> $file") || next;
	    
	    # Try non-blocking wait
	    my $flags = LOCK_EX | LOCK_NB;
#	    $flags |= LOCK_NB; # non-blocking wait?
	    
	    my $lck = flock($this->{lockhandle},$flags);
	    
	    if ($lck) {
		# Got a lock!
		$this->{lockslot} = $n;
		seek($this->{lockhandle},0,0);
		truncate($this->{lockhandle},0);
		print { $this->{lockhandle} } "$$\n";
		$this->{lockhandle}->flush;
		$this->{lockfile} = $file;
		
		$this->log("$$: got lock slot $n\n") if ($this->{debug});
		
		last;
	    }

	    if ($this->{debug} > 5) {
		$this->log("$$: slot $n is busy with code $lck\n");
	    }
	    
	    # Got no lock, so close filehandle
	    close($this->{lockhandle});
	} # for ()

	usleep($this->{lockretrytime});


	if ($this->{debug} > 4) {
	    $this->log("$$: time left: " . ($this->{locktimeout} - (time() - $t0)) . "\n");
	}
    } while (!defined($this->{lockslot}) &&
	     time() - $t0 < $this->{locktimeout}) ;


    if (defined($this->{lockslot})) {
    }
    else {
	$this->log("$$: NO SLOT FOUND within timeout period\n") if ($this->{debug});
    }
}



# Is this lock valid, i. e. good?
sub valid {
    my ($this) = @_;

    return (defined($this->{lockslot}));
}


sub log {
    my ($this,$s) = @_;

    # Logging to logfile or STDERR?
    if (!$this->{logfile} && $this->{debug} > 0) {
	print STDERR "$s";
	return;
    }

    my $fp = new IO::Handle;
    my $file = $this->{logfile};

    if (!open($fp, ">> $file")) {
#	print STDERR "Could not write to logfile!\n";
#	print STDERR "$s";
	return;
    }

    
    # Try non-blocking wait
    my $flags = LOCK_EX;
    # $flags |= LOCK_NB; # non-blocking wait?
    my $lck = flock($fp,$flags);

    if (!$lck) {
	print STDERR "Could not get a lock handle on logfile!\n";
	print STDERR "$s";
	close($fp);
	return;
    }

    seek($fp,0,2);

    print $fp localtime(time) . " " . $s;

    close($fp);
}


sub DESTROY {
    my ($this) = @_;

    $this->log("$$: destroy\n") if ($this->{debug});

    if ($this->{lockfile}) {
	unlink($this->{lockfile});
    }
}


1;




